//
// Created by Falko Alrik Schmidt on 09.09.23.
//

#include "Ulm_SDStorage.h"

Ulm_SDStorage::Ulm_SDStorage() {
    this->csPin = SS;                       // Default SPI SS pin of Arduino.
    this->detectPin = -1;                   // No detect pin.
    this->directory = "/";                  // Root directory on SD card.
    this->dataFileName = String();          // No data file.
    this->redundancyFileName = String();    // No redundancy file.
    this->extremeEnvironmentLogic = false;  // No additional safety measurements.
}

void Ulm_SDStorage::setCsPin(uint8_t cs) {
    this->csPin = cs;
}

void Ulm_SDStorage::setDetectPin(int8_t detect) {
    this->detectPin = detect;
}

void Ulm_SDStorage::setExtremeEnvironmentLogic(bool extremeLogic) {
    this->extremeEnvironmentLogic = extremeLogic;
}

bool Ulm_SDStorage::begin() {
    // Check, if SD card is inserted.
    if (this->detectPin >= 0) {
        pinMode(this->detectPin, INPUT);
        if (digitalRead(this->detectPin) == HIGH) {
            // No SD card inserted!
            return false;
        }
    }

    // Init the SD, open it etc.
    const bool sdInit = SDClass::begin(this->csPin);

    // Create the directory, if not already exists
    if (!SDClass::mkdir(this->directory)) {
        SDClass::end();
        return false;
    }

    // Check data file name for errors
    if (this->dataFileName.length() == 0) {
        // Not set -> error!
        SDClass::end();
        return false;
    }

    // Create data file
    const bool dfInit = this->determineCurrentLogFileName(this->directory, this->dataFileName);
    if(!this->hasRedundancyFile()) {
        // If no redundancy file, then we are done and can check for extreme conditions set.
        if(this->extremeEnvironmentLogic) {
            SDClass::end();
        }
        return sdInit && dfInit;
    }

    // Create redundancy file
    const bool rfInit = this->determineCurrentLogFileName(this->directory, this->redundancyFileName);
    if(this->extremeEnvironmentLogic) {
        SDClass::end();
    }
    return sdInit && dfInit && rfInit;
}

bool Ulm_SDStorage::determineCurrentLogFileName(const String &dir, String &fileName) {
    // Determine the correct file name for data file
    if (fileName.length() == 0) {
        // Not set -> error!
        return false;
    }

    const int numberIndex = fileName.indexOf('0');
    while (SDClass::exists(dir + "/" + fileName)) {
        if (fileName[numberIndex + 2] != '9') {
            fileName[numberIndex + 2]++;
        } else if (fileName[numberIndex + 1] != '9') {
            fileName[numberIndex + 1]++;
            fileName[numberIndex + 2] = '0';
        } else if (fileName[numberIndex] != '9') {
            fileName[numberIndex]++;
            fileName[numberIndex + 1] = '0';
            fileName[numberIndex + 2] = '0';
        } else {
            // All files from 0 to 999 exist -> error!
            return false;
        }
    }
    File f = SDClass::open(dir + "/" + fileName, FILE_WRITE);
    if (f) {
        f.close();
        return true;
    }
    return false;
}

bool Ulm_SDStorage::store(const String &s) {
    return this->store(s.c_str());
}

bool Ulm_SDStorage::store(const char *str) {

    const size_t dataLength = strlen(str);
    const size_t dfWritten = Ulm_SDStorage::storeToFile(str, this->directory, this->dataFileName);

    if (this->hasRedundancyFile()) {
        const size_t redWritten = Ulm_SDStorage::storeToFile(str, this->directory, this->redundancyFileName);
        return dfWritten == dataLength && redWritten == dataLength;
    } else {
        return dfWritten == dataLength;
    }
}


size_t Ulm_SDStorage::storeToFile(const char *str, const String &dir, const String &fileName) {

    String canonicalFileName = dir + "/" + fileName;

    size_t bytesWritten = 0;
    File f = SDClass::open(canonicalFileName, FILE_WRITE);
    if (f) {
        bytesWritten = f.write(str);
        f.flush();
        f.close();
    }
    return bytesWritten;
}


void Ulm_SDStorage::setDirectory(const String &directory) {
    Ulm_SDStorage::directory = directory;
}

void Ulm_SDStorage::setDataFileName(const String &dataFileName) {
    Ulm_SDStorage::dataFileName = dataFileName;
}

void Ulm_SDStorage::setRedundancyFileName(const String &redundancyFileName) {
    Ulm_SDStorage::redundancyFileName = redundancyFileName;
}


Ulm_SDStorage::Ulm_SDStorage_Builder::Ulm_SDStorage_Builder() {
    this->sdStorage = Ulm_SDStorage();
}


// ====================================================================================================================
// ====================================================================================================================
// ====================================================================================================================


Ulm_SDStorage::Ulm_SDStorage_DetectPinBuilder &Ulm_SDStorage::Ulm_SDStorage_Builder::atPin(uint8_t pin) {
    this->sdStorage.setCsPin(pin);
    return *this;
}

Ulm_SDStorage::Ulm_SDStorage_DirectoryBuilder &Ulm_SDStorage::Ulm_SDStorage_Builder::withDetectPin(int8_t detPin) {
    this->sdStorage.setDetectPin(detPin);
    return *this;
}

Ulm_SDStorage::Ulm_SDStorage_DirectoryBuilder &Ulm_SDStorage::Ulm_SDStorage_Builder::withoutDetectPin() {
    this->sdStorage.setDetectPin(-1);
    return *this;
}

Ulm_SDStorage::Ulm_SDStorage_DataFileBuilder &Ulm_SDStorage::Ulm_SDStorage_Builder::atDirectory(const String &dir) {
    const unsigned int dirNameLength = dir.length();
    if (dirNameLength >= 1 && dirNameLength <= 8) {
        // Supported directory name length
        this->sdStorage.setDirectory(dir);
    }
    return *this;
}

Ulm_SDStorage::Ulm_SDStorage_DataFileBuilder &Ulm_SDStorage::Ulm_SDStorage_Builder::atRootDirectory() {
    this->sdStorage.setDirectory("/");
    return *this;
}


String Ulm_SDStorage::determineGenericLogFileName(const String &prefix, Ulm_SDStorage::FileType fileType) {
    String fileName = prefix + "000.";
    switch (fileType) {
        case CSV:   fileName += "csv";  break;
        case LOG:   fileName += "log";  break;
        default:    fileName += "txt";  break;
    }
    return fileName;
}

bool Ulm_SDStorage::hasRedundancyFile() {
    return this->redundancyFileName.length() > 0;
}

Ulm_SDStorage::Ulm_SDStorage_Builder Ulm_SDStorage::builder() {
    return {};
}

Ulm_SDStorage::Ulm_Storage_OptionalArgsBuilder &
Ulm_SDStorage::Ulm_SDStorage_Builder::withDataFile(const String &dataFilePrefix,
                                                   Ulm_SDStorage::FileType fileType) {
    this->sdStorage.setDataFileName(Ulm_SDStorage::determineGenericLogFileName(dataFilePrefix, fileType));
    return *this;
}

Ulm_SDStorage::Ulm_Storage_OptionalArgsBuilder &
Ulm_SDStorage::Ulm_SDStorage_Builder::withRedundancyFile(const String &redFilePrefix,
                                                         Ulm_SDStorage::FileType fileType) {
    this->sdStorage.setRedundancyFileName(Ulm_SDStorage::determineGenericLogFileName(redFilePrefix, fileType));
    return *this;
}


Ulm_SDStorage::Ulm_Storage_OptionalArgsBuilder &
Ulm_SDStorage::Ulm_SDStorage_Builder::withLogicForExtremeEnvironments() {
    this->sdStorage.setExtremeEnvironmentLogic(true);
    return *this;
}

Ulm_SDStorage Ulm_SDStorage::Ulm_SDStorage_Builder::build() {
    return this->sdStorage;
}
